#include "edu_rutgers_winlab_jgnrs_JGNRS.h"
#include "log.h"
#include "object_pool.h"
#include "../cpp/gnrs_cxx.h"
#include "../cpp/guid.h"
#include "../cpp/net_addr.h"


/*
 * Convert a Java string to a c "string"
 */
char *getStrFromJstr(JNIEnv *env, jstring jstr) {
    char *str;
    str = (char*)env->GetStringUTFChars(jstr, NULL);
    if (str == NULL) {
        return NULL; /* OutOfMemoryError already thrown */
    }
    return str;
}

/*
 * From jbyteArray to array of chars
 */
char *getDataFromByteA(JNIEnv *env, jbyteArray data) {
    return (char *)(env->GetByteArrayElements(data, NULL));
}

/*
 * Class:     edu_rutgers_winlab_jgnrs_JGNRS
 * Method:    setNativeGNRS
 * Signature: (Ljava/lang/String;Ljava/lang/String;)I
 */
JNIEXPORT jint JNICALL Java_edu_rutgers_winlab_jgnrs_JGNRS_setNativeGNRS(
        JNIEnv *env, jobject obj, jstring remote, jstring local){

    int gnrsObjIndex = -1;
    if(!objectPoolInited)initObjectPool();
    try{
        //Set up the gnrs object
        string server_addr_s(getStrFromJstr(env, remote));
        string local_addr_s(getStrFromJstr(env, local));
        NetAddr server_addr(NET_ADDR_TYPE_IPV4_PORT, server_addr_s);
        NetAddr local_addr(NET_ADDR_TYPE_IPV4_PORT, local_addr_s);
        /* Alloc new obj to be stored on native side and pass
        * back an index to the local copy
        */
        Gnrs *gnrs = new Gnrs(server_addr, local_addr);
        if((gnrsObjIndex = putGnrsObj(gnrs)) < 0) {
            //capacity reached
            if(gnrsObjIndex == ERR_OBJPOOL_CAPACITY_REACHED) {
                errLog("Reached capacity of GNRS objects");
            } else { 
                errLog("Cannot create GNRS object");
            }
            delete(gnrs);
            gnrsObjIndex = -1;
        }
    } catch(string error){
        errLog(error.c_str());
    } catch(...){
        errLog("Default exception");
    }
    return (jint) gnrsObjIndex;
}

/*
 * Class:     edu_rutgers_winlab_jgnrs_JGNRS
 * Method:    freeNativeGNRS
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_edu_rutgers_winlab_jgnrs_JGNRS_freeNativeGNRS(
        JNIEnv *env, jobject obj, jint gnrsIndex) {

    try{
        if(deleteGnrsObj(gnrsIndex) < 0) {
            errLog("Unable to free, object not found!");
        }
    } catch(...){
        errLog("Default exception");
    }
}

/*
 * Class:     edu_rutgers_winlab_jgnrs_JGNRS
 * Method:    setNativeGUID
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL Java_edu_rutgers_winlab_jgnrs_JGNRS_setNativeGUID(
        JNIEnv *env, jobject obj, jint guid){

    int guidObjIndex = -1;
    if(!objectPoolInited)initObjectPool();
    try{
        Guid *g = new Guid();
        Guid temp = Guid::from_unsigned_int(guid);
        for(int i = 0; i < GUID_BINARY_SIZE; i++){
            g->bytes[i] = temp.bytes[i];
        }
        g->str = temp.str;
        if((guidObjIndex = putGuidObj(g)) < 0) {
            if(guidObjIndex == ERR_OBJPOOL_CAPACITY_REACHED) {
                //over capacity
                errLog("Reached capacity of GUID objects");
            } else {
                errLog("Cannot create GUID object");
            }
            delete g;
            guidObjIndex = -1;
        }
    } catch(string error){
        errLog(error.c_str());
    } catch(...){
        errLog("Default exception");
    }
    return (jint)guidObjIndex;
}

/*
 * Class:     edu_rutgers_winlab_jgnrs_JGNRS
 * Method:    freeNativeGUID
 * Signature: (I)V
 */
JNIEXPORT void JNICALL Java_edu_rutgers_winlab_jgnrs_JGNRS_freeNativeGUID(
        JNIEnv *env, jobject obj, jint guidObjIndex){

    try{
        if(deleteGuidObj(guidObjIndex) < 0) {
            errLog("Unable to free, object not found!");
        }
    } catch(...){
        errLog("Default exception");
    }
}

/*
 * Class:     edu_rutgers_winlab_jgnrs_JGNRS
 * Method:    gnrs_lookup
 * Signature: (IILjava/lang/String;)[I
 */
JNIEXPORT jintArray JNICALL Java_edu_rutgers_winlab_jgnrs_JGNRS_gnrs_1lookup(
        JNIEnv *env, jobject obj, jint gnrsObjIndex, jint guidObjIndex){

	try{
		Guid *nguid = getGuidObj(guidObjIndex);
                if(nguid == NULL){
                    errLog("Invalid param, Guid object not found!");
                    //TODO should return NULL or throw exception
                    return env->NewIntArray(0);
                }
		Gnrs *ngnrs = getGnrsObj(gnrsObjIndex);
                if(ngnrs == NULL){
                    errLog("Invalid param, Gnrs object not found!");
                    //TODO should return NULL or throw exception
                    return env->NewIntArray(0);
                }
		list<NetAddr> lkup_addrs = ngnrs->lookup(*nguid);
		int guids[lkup_addrs.size()];
		int i=0;
		for (list<NetAddr>::const_iterator it = lkup_addrs.begin();
			 it != lkup_addrs.end(); it++) {
			guids[i] = atoi((*it).get_value().c_str());
			i++;
		}
		jintArray out_ints = env->NewIntArray(lkup_addrs.size());
		env->SetIntArrayRegion(out_ints, 0, lkup_addrs.size(), guids);
		return out_ints;
	} catch(string error){
		errLog(error.c_str());
		return env->NewIntArray(0);
	} catch(...){
		errLog("Default exception");
		return  env->NewIntArray(0);
	}
}

/*
 * Class:     edu_rutgers_winlab_jgnrs_JGNRS
 * Method:    gnrs_add
 * Signature: (I[IILjava/lang/String;)I
 */
JNIEXPORT jint JNICALL Java_edu_rutgers_winlab_jgnrs_JGNRS_gnrs_1add(
        JNIEnv *env, jobject obj, jint gnrsObjIndex, jint guidObjIndex, 
        jintArray nas, jint size){

	try{
		Guid *nguid = getGuidObj(guidObjIndex);
                if(nguid == NULL){
                    errLog("Invalid param, Guid object not found!");
                    //TODO should return NULL or throw exception
                    return (jint)-1;
                }
		Gnrs *ngnrs = getGnrsObj(gnrsObjIndex);
                if(ngnrs == NULL){
                    errLog("Invalid param, Gnrs object not found!");
                    //TODO should return NULL or throw exception
                    return (jint)-1;
                }
		char tempAddr[32];
		int *nnas = env->GetIntArrayElements(nas, NULL);
		int s = (int)size;
		list<NetAddr> addrs;
		for(int i = 0; i<s; i++){
                    sprintf(tempAddr, "%d", nnas[i]);
                    NetAddr na(NET_ADDR_TYPE_GUID, tempAddr);
                    addrs.push_back(na);
		}
		ngnrs->add(*nguid, addrs);
		return (jint)1;
	} catch(string error){
		errLog(error.c_str());
		return (jint)-1;
	} catch(...){
		errLog("Default exception");
		return (jint)-1;
	}
}
