/*
 *  Copyright (c) 2010-2013, Rutgers, The State University of New Jersey
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above copyright
 *        notice, this list of conditions and the following disclaimer in the
 *        documentation and/or other materials provided with the distribution.
 *      * Neither the name of the organization(s) stated above nor the
 *        names of its contributors may be used to endorse or promote products
 *        derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
 *  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * @file   mfchunkinfo.h
 * @author wontoniii (bronzino@winlab.rutgers.edu), czhang
 * @date   July, 2011
 * @brief  Temp.
 *
 * Temp.
 */

/*
 * list of packets associated with the chunk
 * |======================================|
 * | room for ether header: 14 bytes      |
 * |--------------------------------------|
 * | room for ip header: 20 bytes (wimax) |
 * |--------------------------------------|
 * | room for hop header: 16 bytes        |
 * |======================================|
 * | room for MF routing header: 16 bytes |
 * |--------------------------------------|
 * | room for transport header: 24 bytes  |  <----- first packet
 * |--------------------------------------|
 * |                                      |
 * |      payload: 1024-16-24 bytes       |
 * |                                      |
 * |======================================|
 *                     |
 *                     |
 *                     v
 * |======================================|
 * | room for ether header: 14 bytes      |
 * |--------------------------------------|
 * | room for ip header: 20 bytes (wimax) |
 * |--------------------------------------|
 * | room for hop header: 16 bytes        |
 * |======================================|
 * |                                      |
 * |                                      |
 * |                                      |
 * |        payload : 1024 bytes          |  <----- intermediate packets
 * |                                      |
 * |                                      |
 * |                                      |
 * |======================================|
 *                     |
 *                     |
 *                     v
 *                  .......
 *                     |
 *                     |
 *                     v
 * |======================================|
 * | room for ether header: 14 bytes      |
 * |--------------------------------------|
 * | room for ip header: 20 bytes (wimax) |
 * |--------------------------------------|
 * | room for hop header: 16 bytes        |
 * |======================================|
 * |                                      |
 * |        payload : < 1024 bytes        |
 * |                                      |  <----- last packet
 * | . . . . . . . . . . . . . . . . . . .|
 * |                                      |
 * |             not used                 |
 * |                                      |
 * |======================================|
 */

#ifndef MF_CHUNKINFO_H_
#define MF_CHUNKINFO_H_

#include <vector>
#include <list>

#include <include/mfflags.h>

#include "mftypes.h"
#include "mflog.h"

using namespace std;

class MF_ChunkInfo {
	u_int mDstNA; //routing, returned by netmanager
	u_int mSrcGUID; //routing, get from a file
	u_int mDstGUID; //routing, get from scheme or mfsendto is used
	u_int mServiceID; //routing
	u_int mChunkID; //transport
	u_int mChunkSize; //transport
	u_int mChunkPktCnt;
	u_int mSrcAppID;
	u_int mDstAppID;
	u_int mMsgID;
	u_short mMsgNum;
	u_short mOffset;
	u_int mHopID; //hop, hold for easy access, need to be updated when re-transmission happen
	bool mIsChunkDelivered; //used for DeviceManager to determine what to do when received a csyn packet
	bool mIsReadyForRelease; //used for DeviceManager to determine what to do when received a csyn packet
	vector<u_char *> mPacketList;
	u_int owner;
	mfflag_t opts;

public:
	
	MF_ChunkInfo();
	MF_ChunkInfo(u_int);
	~MF_ChunkInfo();
	u_int getPacketCount();
	u_int getChunkPktCnt();
	void putChunkPktCnt(u_int);
	void reset();
	void putSrcGUID(u_int);
	u_int getSrcGUID();
	void putDstGUID(u_int);
	u_int getDstGUID();
	void putServiceID(u_int);
	u_int getServiceID();
	void putChunkID(u_int);
	u_int getChunkID();
	void putChunkSize(u_int);
	u_int getChunkSize();
	void putDstNA(u_int);
	void putHopID(u_int);
	u_int getHopID();
	void putSrcTID(u_int);
	u_int getSrcTID();
	void putDstTID(u_int);
	u_int getDstTID();
	void putMsgID(u_int);
	u_int getMsgID();
	void putMsgNum(u_short);
	u_short getMsgNum();
	void putOffset(u_short);
	u_short getOffset();
	bool isDelivered();
	void setDelivered(bool val);
	bool isReleased();
	void setReleased(bool val);
	void putPktList(vector<u_char*> *);
	vector<u_char *> *getPacketList();
	u_int getOwner();
	void setOwner(u_int);
	mfflag_t getOpts();
	bool isMultihoming();
	bool isMulticast();
	bool isBroadcast();
	bool isAnycast();
	bool isContentRequest();
	bool isContentResponse();
	void setOpts(mfflag_t);
};

#endif /* MF_CHUNKINFO_H_ */
